@@ -11,7 +11,7 @@ from typing import Mapping
 from typing import Type
 
 from .builtin import DictLoader
-from .builtin import register_standard_tags_and_filters
+from .builtin import register_default_tags_and_filters
 from .exceptions import LiquidError
 from .lexer import Lexer
 from .parser import Parser
@@ -48,6 +48,9 @@ class Environment:
         default_trim: The automatic whitespace stripping mode to use. This mode can then
             be overridden by template authors per Liquid tag using whitespace control
             symbols (`-`, `+`, `~`).
+        validate_filter_arguments: If `True`, class-based filters that define a
+            `validate()` method will have their arguments validated as each template is
+            parsed.
     """
 
     context_depth_limit: ClassVar[int] = 30
@@ -85,20 +88,27 @@ class Environment:
         auto_escape: bool = False,
         undefined: Type[Undefined] = Undefined,
         default_trim: WhitespaceControl = WhitespaceControl.PLUS,
+        validate_filter_arguments: bool = True,
     ) -> None:
         self.loader = loader or DictLoader({})
         self.globals = globals or {}
         self.auto_escape = auto_escape
         self.undefined = undefined
+        self.validate_filter_arguments = validate_filter_arguments
 
         self.default_trim: WhitespaceControl = (
             WhitespaceControl.PLUS
             if default_trim == WhitespaceControl.DEFAULT
             else default_trim
         )
+        """The default whitespace trimming mode."""
 
         self.filters: dict[str, Callable[..., Any]] = {}
+        """The environment's filter register, mapping filter names to callables."""
+
         self.tags: dict[str, Tag] = {}
+        """The environment's tag register, mapping tag names to instances of `Tag`."""
+
         self.setup_tags_and_filters()
         self.parser = Parser(self)
 
@@ -108,7 +118,7 @@ class Environment:
         This is called once when initializing an environment. Override this method
         in your custom environments.
         """
-        register_standard_tags_and_filters(self)
+        register_default_tags_and_filters(self)
 
     def tokenize(self, source: str) -> list[TokenT]:
         """Scan Liquid template _source_ and return a list of Markup objects."""
